#include "ObjModel.h"

bool ObjModel::Initialize(char* filename)
{
	ifstream fin;
	bool result;
	int vertexCount, normalCount, textureCount, faceCount, materialCount;

	//Check that the file will open correctly
	fin.open(filename);

	if (fin.good())
	{
		result = ReadFileCount(filename, vertexCount, textureCount, normalCount, faceCount);
		if (!result)
		{
			MessageBox(NULL, "There was a problem reading the file count.", "Error", MB_OK);
			return false;
		}

		result = LoadDataStructures(filename, vertexCount, textureCount, normalCount, faceCount);
		if (!result)
		{
			MessageBox(NULL, "There was a problem reading the data from the file.", "Error", MB_OK);
			return false;
		}
	}
	else
	{
		MessageBox(NULL, "There was a problem opening the .obj file.", "Error", MB_OK);
		return false;
	}

	return true;
}

bool ObjModel::ReadFileCount(char* modelFilename, int& vertexCount, int& textureCount, int& normalCount, int& faceCount)
{
	ifstream fin;
	char input;

	vertexCount = 0;
	textureCount = 0;
	normalCount = 0;
	faceCount = 0;

	m_meshCount = 0;

	fin.open(modelFilename);

	if (fin.fail() == true)
	{
		return false;
	}

	fin.get(input);
	while (!fin.eof())
	{
		//if the line starts with 'v' then increment the correct count
		if (input == 'v')
		{
			fin.get(input);
			if (input == ' ') { vertexCount++; }
			if (input == 't') { textureCount++; }
			if (input == 'n') { normalCount++; }
		}

		//if the line starts with 'f' increment the face count
		if (input == 'f')
		{
			fin.get(input);
			if (input == ' '){ faceCount++; }
		}

		//if the line starts with 'u' increment the mesh count
		if (input == 'u')
		{
			if (m_meshCount > 0)
			{
				facesPerMesh.push_back(faceCount);
				faceCount = 0;
			}
			m_meshCount++;
		}

		//Otherwise read the rest of the line
		while (input != '\n' && !fin.eof())
		{
			fin.get(input);
		}

		fin.get(input);
	}

	fin.close();

	return true;
}

bool ObjModel::LoadDataStructures(char* filename, int vertexCount, int textureCount, int normalCount, int faceCount)
{
	LoadVertex *vertices, *texcoords, *normals;
	LoadMesh *tempMeshes;
	LoadFace *faces;
	ifstream fin;
	int vertexIndex, texcoordIndex, normalIndex, faceIndex, vIndex, tIndex, nIndex, meshIndex, meshFaceIndex;
	char input, input2, materialFilename, materialName;
	int vertexNum;

	vertices = new LoadVertex[vertexCount];
	if (!vertices)
	{
		return false;
	}

	texcoords = new LoadVertex[textureCount];
	if (!texcoords)
	{
		return false;
	}

	normals = new LoadVertex[normalCount];
	if (!normals)
	{
		return false;
	}

	faces = new LoadFace[faceCount];
	if (!faces)
	{
		return false;
	}

	tempMeshes = new LoadMesh[m_meshCount];
	if (!tempMeshes)
	{
		return false;
	}

	for (int i = 0; i < m_meshCount; i++)
	{
		tempMeshes[i].numFaces = facesPerMesh[i];
		tempMeshes[i].m_faces = new LoadFace[facesPerMesh[i]];
	}

	vertexIndex = 0;
	texcoordIndex = 0;
	normalIndex = 0;
	faceIndex = 0;
	meshIndex = -1;
	meshFaceIndex = 0;

	fin.open(filename);

	if (fin.fail() == true)
	{
		return false;
	}

	fin.get(input);

	while (!fin.eof())
	{
		if (input == 'v')
		{
			fin.get(input);

			//Read vertex data
			if (input == ' ')
			{
				fin >> vertices[vertexIndex].x >> vertices[vertexIndex].y >> vertices[vertexIndex].z;
				vertexIndex++;
			}

			//Read in the texture uv coordinates
			if (input == 't')
			{
				fin >> texcoords[texcoordIndex].x >> texcoords[texcoordIndex].y;
				texcoordIndex++;
			}

			//Read in normals
			if (input == 'n')
			{
				fin >> normals[normalIndex].x >> normals[normalIndex].y >> normals[normalIndex].z;
				normalIndex++;
			}
		}

		//Read in the faces
		if (input == 'f')
		{
			fin.get(input);
			if (input == ' ')
			{
				//Read in the face data
				fin >> faces[faceIndex].vIndex1 >> input2 >> faces[faceIndex].tIndex1 >> input2 >> faces[faceIndex].nIndex1
					>> faces[faceIndex].vIndex2 >> input2 >> faces[faceIndex].tIndex2 >> input2 >> faces[faceIndex].nIndex2
					>> faces[faceIndex].vIndex3 >> input2 >> faces[faceIndex].tIndex3 >> input2 >> faces[faceIndex].nIndex3;
				if (m_meshCount > 0)
				{
					tempMeshes[meshIndex].m_faces[meshFaceIndex] = faces[faceIndex];
					meshFaceIndex++;
				}
				faceIndex++;

			}
		}

		//if the line starts with 'm' assume that the material file is being read in
		if (input == 'm')
		{
			while (input != ' ')
			{
				fin.get(input);
			}
			fin.getline(&materialFilename, 100);

		}

		//if the line starts with 'u' assume that the material name is being read in and that a new mesh is being read in
		if (input == 'u')
		{
			while (input != ' ')
			{
				fin.get(input);
			}
			fin.getline(&materialName, 100);
			meshIndex++;
			meshFaceIndex = 0;
		}

		while (input != '\n' && !fin.eof())
		{
			fin.get(input);
		}

		fin.get(input);
	}

	fin.close();

	//Initialize the vertex array
	m_vertexCount = faceCount * 3;

	m_vertices = new VertexType[m_vertexCount];
	if (!m_vertices)
	{
		return false;
	}

	//Begin writing the data into the model
	for (int i = 0; i < faceCount; i++)
	{
		vertexNum = i * 3;

		vIndex = faces[i].vIndex1 - 1;
		tIndex = faces[i].tIndex1 - 1;
		nIndex = faces[i].nIndex1 - 1;

		m_vertices[vertexNum].x = vertices[vIndex].x;
		m_vertices[vertexNum].y = vertices[vIndex].y;
		m_vertices[vertexNum].z = vertices[vIndex].z;

		m_vertices[vertexNum].tu = texcoords[tIndex].x;
		m_vertices[vertexNum].tv = 1.0f - texcoords[tIndex].y;

		m_vertices[vertexNum].nx = normals[nIndex].x;
		m_vertices[vertexNum].ny = normals[nIndex].y;
		m_vertices[vertexNum].nz = normals[nIndex].z;

		vertexNum++;
		vIndex = faces[i].vIndex2 - 1;
		tIndex = faces[i].tIndex2 - 1;
		nIndex = faces[i].nIndex2 - 1;

		m_vertices[vertexNum].x = vertices[vIndex].x;
		m_vertices[vertexNum].y = vertices[vIndex].y;
		m_vertices[vertexNum].z = vertices[vIndex].z;

		m_vertices[vertexNum].tu = texcoords[tIndex].x;
		m_vertices[vertexNum].tv = 1.0f - texcoords[tIndex].y;

		m_vertices[vertexNum].nx = normals[nIndex].x;
		m_vertices[vertexNum].ny = normals[nIndex].y;
		m_vertices[vertexNum].nz = normals[nIndex].z;

		vertexNum++;
		vIndex = faces[i].vIndex3 - 1;
		tIndex = faces[i].tIndex3 - 1;
		nIndex = faces[i].nIndex3 - 1;

		m_vertices[vertexNum].x = vertices[vIndex].x;
		m_vertices[vertexNum].y = vertices[vIndex].y;
		m_vertices[vertexNum].z = vertices[vIndex].z;

		m_vertices[vertexNum].tu = texcoords[tIndex].x;
		m_vertices[vertexNum].tv = 1.0f - texcoords[tIndex].y;

		m_vertices[vertexNum].nx = normals[nIndex].x;
		m_vertices[vertexNum].ny = normals[nIndex].y;
		m_vertices[vertexNum].nz = normals[nIndex].z;
	}

	vertex_array = new float[m_vertexCount * 3];
	tex_array = new float[m_vertexCount * 2];
	normal_array = new float[m_vertexCount * 3];

	int vertIndex = 0, texIndex = 0, normalsIndex = 0;

	for (int i = 0; i < m_vertexCount; i++)
	{
		vertex_array[vertIndex] = m_vertices[i].x;
		vertIndex++;
		vertex_array[vertIndex] = m_vertices[i].y;
		vertIndex++;
		vertex_array[vertIndex] = m_vertices[i].z;
		vertIndex++;

		tex_array[texIndex] = m_vertices[i].tu;
		texIndex++;
		tex_array[texIndex] = m_vertices[i].tv;
		texIndex++;

		normal_array[normalsIndex] = m_vertices[i].nx;
		normalsIndex++;
		normal_array[normalsIndex] = m_vertices[i].ny;
		normalsIndex++;
		normal_array[normalsIndex] = m_vertices[i].nz;
		normalsIndex++;
	}

	if (vertices)
	{
		delete[] vertices;
		vertices = nullptr;
	}

	if (texcoords)
	{
		delete[] texcoords;
		texcoords = nullptr;
	}

	if (normals)
	{
		delete[] normals;
		normals = nullptr;
	}

	if (faces)
	{
		delete[] faces;
		faces = nullptr;
	}

	return true;
}

void ObjModel::Render()
{
	glBindTexture(GL_TEXTURE_2D, GetTexture());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	float x, y, z;

	glPushMatrix();
	GetTranslation(x, y, z);
	glTranslatef(x, y, z);
	GetRotation(x, y, z);
	glRotatef(z, 0, 0, 1);
	glRotatef(y, 0, 1, 0);
	glRotatef(x, 1, 0, 0);
	GetScaling(x, y, z);
	glScalef(x, y, z);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);


	glVertexPointer(3, GL_FLOAT, 0, vertex_array);
	glNormalPointer(GL_FLOAT, 0, normal_array);
	glTexCoordPointer(2, GL_FLOAT, 0, tex_array);

	glDrawArrays(GL_TRIANGLES, 0, m_vertexCount);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	glDisable(GL_BLEND);
	glPopMatrix();
}