#include "TerrainModel.h"

bool TerrainModel::Initialize(char* filename)
{
	LoadRawHeightMap(filename);

	SetTerrainCoordinates();

	CalculateNormals();

	BuildTerrainModel();
	
	InitializeQuadTree();

	return true;
}

bool TerrainModel::LoadHeightMap(char* filename)
{
	FILE* filePtr;
	int error;
	unsigned int count;
	BITMAPFILEHEADER bitmapFileHeader;
	BITMAPINFOHEADER bitmapInfoHeader;
	int imageSize, k, index;
	unsigned char* bitmapImage;
	unsigned char height;
	int terrainWidth, terrainHeight;
	float* heightMapData;

	// Open the height map file in binary.
	error = fopen_s(&filePtr, filename, "rb");
	if (error != 0)
	{
		return false;
	}

	// Read in the file header.
	count = fread(&bitmapFileHeader, sizeof(BITMAPFILEHEADER), 1, filePtr);
	if (count != 1)
	{
		return false;
	}

	// Read in the bitmap info header.
	count = fread(&bitmapInfoHeader, sizeof(BITMAPINFOHEADER), 1, filePtr);
	if (count != 1)
	{
		return false;
	}

	terrainWidth = bitmapInfoHeader.biWidth;
	terrainHeight = bitmapInfoHeader.biHeight;

	m_terrainWidth = terrainWidth;
	m_terrainHeight = terrainHeight;

	imageSize = terrainWidth * terrainHeight * 3;

	bitmapImage = new unsigned char[imageSize];
	if (!bitmapImage)
	{
		return false;
	}

	//Move to the beginning of the image data
	fseek(filePtr, bitmapFileHeader.bfOffBits, SEEK_SET);

	//Read in the bitmap data
	count = fread(bitmapImage, 1, imageSize, filePtr);
	if (count != imageSize)
	{
		return false;
	}

	error = fclose(filePtr);
	if (error != 0)
	{
		return false;
	}

	heightMapData = new float[terrainWidth * terrainHeight];
	if (!heightMapData)
	{
		return false;
	}

	k = 0;

	for (int j = 0; j < terrainHeight; j++)
	{
		for (int i = 0; i < terrainWidth; i++)
		{
			height = bitmapImage[k];

			index = (terrainHeight * j) + i;

			heightMapData[index] = (float)height;

			//heightMapData[index] /= 256.0f;

			k += 3;
		}
	}

	for (int i = 0; i < m_vertexCount; i++)
	{
		int map_index;
		float mapX, mapZ;

		if (m_vertices[i].x == 0.5f)
		{
			mapX = 0.5f - (1.0f / m_terrainWidth);
		}
		else

		{
			mapX = m_vertices[i].x;
		}

		if (m_vertices[i].z == 0.5f)
		{
			mapZ = 0.5f - (1.0f / m_terrainHeight);
		}
		else

		{
			mapZ = m_vertices[i].z;
		}

		map_index = (((mapX + 0.5f) / (1.0f / (float)m_terrainWidth)) * (float)m_terrainHeight) + ((mapZ + 0.5f) / (1.0f / (float)m_terrainHeight));
		m_vertices[i].y = (heightMapData[map_index] / 256.0f);
	}


	delete[] bitmapImage;
	bitmapImage = nullptr;

	return true;
}

void TerrainModel::Shutdown()
{
	if (m_parentNode)
	{
		ReleaseNode(m_parentNode);
		delete m_parentNode;
		m_parentNode = nullptr;
	}
}

void TerrainModel::Render()
{
	glBindTexture(GL_TEXTURE_2D, GetTexture());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	RenderNode(m_parentNode);
}

bool TerrainModel::LoadRawHeightMap(char* filename)
{
	int error, index;
	bool result;
	FILE* filePtr;
	unsigned long long imageSize, count;
	unsigned short* rawImage;
	float* heightMapData;

	m_terrainWidth = 256;
	m_terrainHeight = 256;

	m_heightMap = new VertexType[m_terrainWidth * m_terrainHeight];
	if (!m_heightMap)
	{
		return false;
	}

	imageSize = m_terrainWidth * m_terrainHeight;

	error = fopen_s(&filePtr, filename, "rb");
	if (error != 0)
	{
		return false;
	}

	rawImage = new unsigned short[imageSize];
	if (!rawImage)
	{
		return false;
	}

	count = fread(rawImage, sizeof(unsigned short), imageSize, filePtr);
	if (count != imageSize)
	{
		return false;
	}

	error = fclose(filePtr);
	if (error != 0)
	{
		return false;
	}

	for (int j = 0; j < m_terrainHeight; j++)
	{
		for (int i = 0; i < m_terrainWidth; i++)
		{
			index = (m_terrainWidth * j) + i;

			m_heightMap[index].y = (float)rawImage[index];
		}
	}

	delete[] rawImage;
	rawImage = nullptr;

	return true;
}

void TerrainModel::SetTerrainCoordinates()
{
	int index;

	for (int j = 0; j < m_terrainHeight; j++)
	{
		for (int i = 0; i < m_terrainWidth; i++)
		{
			index = (m_terrainWidth * j) + i;
			
			m_heightMap[index].x = (float)i;
			m_heightMap[index].z = -(float)j;

			//centre the heightmap around the origin
			m_heightMap[index].x -= (float)(m_terrainWidth - 1) / 2;
			m_heightMap[index].z += (float)(m_terrainHeight - 1)/2;

			m_heightMap[index].y /= 300.0f;
		}
	}
}

bool TerrainModel::BuildTerrainModel()
{
	int index, index1, index2, index3, index4;

	m_vertexCount = (m_terrainHeight - 1) *(m_terrainWidth - 1) * 6;

	m_vertices = new VertexType[m_vertexCount];
	if (!m_vertices)
	{
		return false;
	}

	index = 0;

	for (int j = 0; j < (m_terrainHeight - 1); j++)
	{
		for (int i = 0; i < (m_terrainWidth - 1); i++)
		{
			index1 = (m_terrainWidth * j) + i;
			index2 = (m_terrainWidth* j) + (i + 1);
			index3 = (m_terrainWidth * (j + 1)) + i;
			index4 = (m_terrainWidth * (j + 1)) + (i + 1);

			m_vertices[index].x = m_heightMap[index1].x;
			m_vertices[index].y = m_heightMap[index1].y;
			m_vertices[index].z = m_heightMap[index1].z;
			m_vertices[index].tu = 0.0f;
			m_vertices[index].tv = 0.0f;
			m_vertices[index].nx = m_heightMap[index1].nx;
			m_vertices[index].ny = m_heightMap[index1].ny;
			m_vertices[index].nz = m_heightMap[index1].nz;
			index++;

			m_vertices[index].x = m_heightMap[index2].x;
			m_vertices[index].y = m_heightMap[index2].y;
			m_vertices[index].z = m_heightMap[index2].z;
			m_vertices[index].tu = 1.0f;
			m_vertices[index].tv = 0.0f;
			m_vertices[index].nx = m_heightMap[index2].nx;
			m_vertices[index].ny = m_heightMap[index2].ny;
			m_vertices[index].nz = m_heightMap[index2].nz;
			index++;

			m_vertices[index].x = m_heightMap[index4].x;
			m_vertices[index].y = m_heightMap[index4].y;
			m_vertices[index].z = m_heightMap[index4].z;
			m_vertices[index].tu = 1.0f;
			m_vertices[index].tv = 1.0f;
			m_vertices[index].nx = m_heightMap[index4].nx;
			m_vertices[index].ny = m_heightMap[index4].ny;
			m_vertices[index].nz = m_heightMap[index4].nz;
			index++;

			m_vertices[index].x = m_heightMap[index1].x;
			m_vertices[index].y = m_heightMap[index1].y;
			m_vertices[index].z = m_heightMap[index1].z;
			m_vertices[index].tu = 0.0f;
			m_vertices[index].tv = 0.0f;
			m_vertices[index].nx = m_heightMap[index1].nx;
			m_vertices[index].ny = m_heightMap[index1].ny;
			m_vertices[index].nz = m_heightMap[index1].nz;
			index++;

			m_vertices[index].x = m_heightMap[index4].x;
			m_vertices[index].y = m_heightMap[index4].y;
			m_vertices[index].z = m_heightMap[index4].z;
			m_vertices[index].tu = 1.0f;
			m_vertices[index].tv = 1.0f;
			m_vertices[index].nx = m_heightMap[index4].nx;
			m_vertices[index].ny = m_heightMap[index4].ny;
			m_vertices[index].nz = m_heightMap[index4].nz;
			index++;

			m_vertices[index].x = m_heightMap[index3].x;
			m_vertices[index].y = m_heightMap[index3].y;
			m_vertices[index].z = m_heightMap[index3].z;
			m_vertices[index].tu = 0.0f;
			m_vertices[index].tv = 1.0f;
			m_vertices[index].nx = m_heightMap[index3].nx;
			m_vertices[index].ny = m_heightMap[index3].ny;
			m_vertices[index].nz = m_heightMap[index3].nz;
			index++;

		}
	}

	return true;
}

void TerrainModel::InitializeQuadTree()
{
	int vertexCount;
	float centerX, centerZ, width;

	vertexCount = m_vertexCount;

	m_triangleCount = vertexCount / 3;

	CalculateMeshDimensions(vertexCount, centerX, centerZ, width);

	m_parentNode = new NodeType;

	CreateTreeNode(m_parentNode, centerX, centerZ, width);

	delete[] m_vertices;
	m_vertices = nullptr;
}

void TerrainModel::CalculateMeshDimensions(int vertexCount, float& centerX, float& centerZ, float& meshWidth)
{
	float maxWidth, maxDepth, minWidth, minDepth, width, depth, maxX, maxZ;

	centerX = 0.0f;
	centerZ = 0.0f;

	for (int i = 0; i < vertexCount; i++)
	{
		centerX += m_vertices[i].x;
		centerZ += m_vertices[i].z;
	}

	centerX = centerX / (float)vertexCount;
	centerZ = centerZ / (float)vertexCount;

	maxWidth = 0.0f;
	maxDepth = 0.0f;

	minWidth = fabsf(m_vertices[0].x - centerX);
	minDepth = fabsf(m_vertices[0].y - centerZ);

	for (int i = 0; i < vertexCount; i++)
	{
		width = fabsf(m_vertices[i].x - centerX);
		depth = fabsf(m_vertices[i].z - centerZ);

		if (width > maxWidth){ maxWidth = width; }
		if (depth > maxDepth){ maxDepth = depth; }
		if (width < minWidth){ minWidth = width; }
		if (depth < minDepth){ minDepth = depth; }
	}

	maxX = (float)std::max(fabs(minWidth), fabs(maxWidth));
	maxZ = (float)std::max(fabs(minDepth), fabs(maxDepth));

	meshWidth = max(maxX, maxZ) * 2.0f;
}

void TerrainModel::CreateTreeNode(NodeType* node, float positionX, float positionZ, float width)
{
	int numTriangles, count, vertexCount, index, vertexIndex;
	float offsetX, offsetZ;
	unsigned long* indices;
	bool result;

	node->positionX = positionX;
	node->positionZ = positionZ;
	node->width = width;

	node->triangleCount = 0;

	node->vertices = nullptr;

	node->nodes[0] = nullptr;
	node->nodes[1] = nullptr;
	node->nodes[2] = nullptr;
	node->nodes[3] = nullptr;

	numTriangles = CountTriangles(positionX, positionZ, width);

	if (numTriangles == 0)
	{
		return;
	}

	if (numTriangles > MAX_TRIANGLES)
	{
		for (int i = 0; i < 4; i++)
		{
			offsetX = (((i % 2) < 1) ? -1.0f : 1.0f) * (width / 4.0f);
			offsetZ = (((i % 4) < 2) ? -1.0f : 1.0f) * (width / 4.0f);

			count = CountTriangles((positionX + offsetX), (positionZ + offsetZ), (width / 2.0f));
			if (count > 0)
			{
				node->nodes[i] = new NodeType;

				CreateTreeNode(node->nodes[i], (positionX + offsetX), (positionZ + offsetZ), (width/2.0f));
			}
		}

		return;
	}

	node->triangleCount = numTriangles;

	vertexCount = numTriangles * 3;

	node->vertices = new VertexType[vertexCount];

	thread threads[THREAD_COUNT];
	NodeThreadArgs threadArgs[THREAD_COUNT];

	index = 0;
	
	for (int i = 0; i < THREAD_COUNT; i++)
	{
		int start_index = ((float)m_triangleCount / (float)THREAD_COUNT) * (float)i;
		int end_index = ((float)m_triangleCount / (float)THREAD_COUNT) * (float)(i + 1);

		threadArgs[i] = NodeThreadArgs{ start_index, end_index, positionX, positionZ, width, node, &index};

		threads[i] = thread(&TerrainModel::CalculateNodeThreaded, this, threadArgs[i]);
	}

	for (int i = 0; i < THREAD_COUNT; i++)
	{
		threads[i].join();
	}
	
	/*
	for (int i = 0; i < m_triangleCount; i++)
	{
		result = IsTriangleContained(i, positionX, positionZ, width);
		if (result)
		{
			vertexIndex = i * 3;

			node->vertices[index].x = m_vertices[vertexIndex].x;
			node->vertices[index].y = m_vertices[vertexIndex].y;
			node->vertices[index].z = m_vertices[vertexIndex].z;
			node->vertices[index].tu = m_vertices[vertexIndex].tu;
			node->vertices[index].tv = m_vertices[vertexIndex].tv;
			node->vertices[index].nx = m_vertices[vertexIndex].nx;
			node->vertices[index].ny = m_vertices[vertexIndex].ny;
			node->vertices[index].nz = m_vertices[vertexIndex].nz;
			index++;

			vertexIndex++;
			node->vertices[index].x = m_vertices[vertexIndex].x;
			node->vertices[index].y = m_vertices[vertexIndex].y;
			node->vertices[index].z = m_vertices[vertexIndex].z;
			node->vertices[index].tu = m_vertices[vertexIndex].tu;
			node->vertices[index].tv = m_vertices[vertexIndex].tv;
			node->vertices[index].nx = m_vertices[vertexIndex].nx;
			node->vertices[index].ny = m_vertices[vertexIndex].ny;
			node->vertices[index].nz = m_vertices[vertexIndex].nz;
			index++;

			vertexIndex++;
			node->vertices[index].x = m_vertices[vertexIndex].x;
			node->vertices[index].y = m_vertices[vertexIndex].y;
			node->vertices[index].z = m_vertices[vertexIndex].z;
			node->vertices[index].tu = m_vertices[vertexIndex].tu;
			node->vertices[index].tv = m_vertices[vertexIndex].tv;
			node->vertices[index].nx = m_vertices[vertexIndex].nx;
			node->vertices[index].ny = m_vertices[vertexIndex].ny;
			node->vertices[index].nz = m_vertices[vertexIndex].nz;
			index++;
		}
	}
	*/
}

void TerrainModel::CalculateNodeThreaded(NodeThreadArgs args)
{
	bool result;
	int vertexIndex;

	for (int i = args.start_index; i < args.end_index; i++)
	{
		result = IsTriangleContained(i, args.positionX, args.positionZ, args.width);
		if (result)
		{
			vertexIndex = i * 3;

			std::lock_guard<mutex> lock(count_mutex);
			args.node->vertices[*(args.index)].x = m_vertices[vertexIndex].x;
			args.node->vertices[*(args.index)].y = m_vertices[vertexIndex].y;
			args.node->vertices[*(args.index)].z = m_vertices[vertexIndex].z;
			args.node->vertices[*(args.index)].tu = m_vertices[vertexIndex].tu;
			args.node->vertices[*(args.index)].tv = m_vertices[vertexIndex].tv;
			args.node->vertices[*(args.index)].nx = m_vertices[vertexIndex].nx;
			args.node->vertices[*(args.index)].ny = m_vertices[vertexIndex].ny;
			args.node->vertices[*(args.index)].nz = m_vertices[vertexIndex].nz;
			*(args.index) += 1;
			
			vertexIndex++;
			args.node->vertices[*(args.index)].x = m_vertices[vertexIndex].x;
			args.node->vertices[*(args.index)].y = m_vertices[vertexIndex].y;
			args.node->vertices[*(args.index)].z = m_vertices[vertexIndex].z;
			args.node->vertices[*(args.index)].tu = m_vertices[vertexIndex].tu;
			args.node->vertices[*(args.index)].tv = m_vertices[vertexIndex].tv;
			args.node->vertices[*(args.index)].nx = m_vertices[vertexIndex].nx;
			args.node->vertices[*(args.index)].ny = m_vertices[vertexIndex].ny;
			args.node->vertices[*(args.index)].nz = m_vertices[vertexIndex].nz;
			*(args.index) += 1;

			vertexIndex++;
			args.node->vertices[*(args.index)].x = m_vertices[vertexIndex].x;
			args.node->vertices[*(args.index)].y = m_vertices[vertexIndex].y;
			args.node->vertices[*(args.index)].z = m_vertices[vertexIndex].z;
			args.node->vertices[*(args.index)].tu = m_vertices[vertexIndex].tu;
			args.node->vertices[*(args.index)].tv = m_vertices[vertexIndex].tv;
			args.node->vertices[*(args.index)].nx = m_vertices[vertexIndex].nx;
			args.node->vertices[*(args.index)].ny = m_vertices[vertexIndex].ny;
			args.node->vertices[*(args.index)].nz = m_vertices[vertexIndex].nz;
			*(args.index) += 1;
		}
	}
}

int TerrainModel::CountTriangles(float positionX, float positionZ, float width)
{
	int tri_count;
	bool result;

	thread threads[THREAD_COUNT];
	TriangleThreadArgs threadArgs[THREAD_COUNT];

	tri_count = 0;
	
	for (int i = 0; i < THREAD_COUNT; i++)
	{
		int start_index = ((float)m_triangleCount / (float)THREAD_COUNT) * (float)i;
		int end_index = ((float)m_triangleCount / (float)THREAD_COUNT) * (float)(i+1);

		threadArgs[i] = TriangleThreadArgs{ start_index, end_index, positionX, positionZ, width, &tri_count };

		threads[i] = thread(&TerrainModel::IsTriangleContainedThreaded, this, threadArgs[i]);
	}

	for (int i = 0; i < THREAD_COUNT; i++)
	{
		threads[i].join();
	}
	
	/*
	for (int i = 0; i < m_triangleCount; i++)
	{
		result = IsTriangleContained(i, positionX, positionZ, width);
		if (result)
		{
			tri_count++;
		}
	}
	*/
	return tri_count;
}

bool TerrainModel::IsTriangleContained(int index, float positionX, float positionZ, float width)
{
	float radius;
	int vertexIndex;
	float x1, z1, x2, z2, x3, z3;
	float minimumX, maximumX, minimumZ, maximumZ;

	radius = width / 2.0f;

	vertexIndex = index * 3;

	x1 = m_vertices[vertexIndex].x;
	z1 = m_vertices[vertexIndex].z;
	vertexIndex++;

	x2 = m_vertices[vertexIndex].x;
	z2 = m_vertices[vertexIndex].z;
	vertexIndex++;

	x3 = m_vertices[vertexIndex].x;
	z3 = m_vertices[vertexIndex].z;

	minimumX = min(x1, min(x2, x3));
	if (minimumX > (positionX + radius))
	{
		return false;
	}

	maximumX = max(x1, max(x2, x3));
	if (maximumX < (positionX - radius))
	{
		return false;
	}

	minimumZ = min(z1, min(z2, z3));
	if (minimumZ > (positionZ + radius))
	{
		return false;
	}

	maximumZ = max(z1, max(z2, z3));
	if (maximumZ < (positionZ - radius))
	{
		return false;
	}

	return true;
}

void TerrainModel::IsTriangleContainedThreaded(TriangleThreadArgs args)
{
	float radius;
	int vertexIndex;
	float x1, z1, x2, z2, x3, z3;
	float minimumX, maximumX, minimumZ, maximumZ;

	radius = args.width / 2.0f;

	for (int i = args.start_index; i < args.end_index; i++)
	{
		vertexIndex = i * 3;

		x1 = m_vertices[vertexIndex].x;
		z1 = m_vertices[vertexIndex].z;
		vertexIndex++;

		x2 = m_vertices[vertexIndex].x;
		z2 = m_vertices[vertexIndex].z;
		vertexIndex++;

		x3 = m_vertices[vertexIndex].x;
		z3 = m_vertices[vertexIndex].z;

		minimumX = min(x1, min(x2, x3));
		if (minimumX > (args.positionX + radius))
		{
			continue;
		}

		maximumX = max(x1, max(x2, x3));
		if (maximumX < (args.positionX - radius))
		{
			continue;
		}

		minimumZ = min(z1, min(z2, z3));
		if (minimumZ >(args.positionZ + radius))
		{
			continue;
		}

		maximumZ = max(z1, max(z2, z3));
		if (maximumZ < (args.positionZ - radius))
		{
			continue;
		}

		std::lock_guard<mutex> lock(count_mutex);
		*(args.count) += 1;
	}
}

void TerrainModel::ReleaseNode(NodeType* node)
{
	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i] != nullptr)
		{
			ReleaseNode(node->nodes[i]);
		}
	}

	if (node->vertices)
	{
		delete[] node->vertices;
		node->vertices = nullptr;
	}

	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i])
		{
			delete node->nodes[i];
			node->nodes[i] = nullptr;
		}
	}
}

void TerrainModel::RenderNode(NodeType* node)
{
	int count;
	float x, y, z;

	count = 0;
	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i] != nullptr)
		{
			count++;
			RenderNode(node->nodes[i]);
		}
	}

	if (count != 0)
	{
		return;
	}

	glPushMatrix();
	GetScaling(x, y, z);
	glScalef(x, y, z);
	GetTranslation(x, y, z);
	glTranslatef(x, y, z);
	GetRotation(x, y, z);
	glRotatef(z, 0, 0, 1);
	glRotatef(y, 0, 1, 0);
	glRotatef(x, 1, 0, 0);

	glBegin(GL_TRIANGLES);


	for (int i = 0; i < (node->triangleCount * 3); i++)
	{
		glNormal3f(node->vertices[i].nx, node->vertices[i].ny, node->vertices[i].nz);
		glTexCoord2f(node->vertices[i].tu, node->vertices[i].tv);
		glVertex3f(node->vertices[i].x, node->vertices[i].y, node->vertices[i].z);
	}

	glEnd();
	glPopMatrix();
}

bool TerrainModel::GetHeightAtPosition(float positionX, float positionZ, float& height)
{
	float meshMinX, meshMaxX, meshMinZ, meshMaxZ;

	meshMinX = (m_parentNode->positionX - (m_parentNode->width / 2.0f)) * m_scale.x;
	meshMaxX = (m_parentNode->positionX + (m_parentNode->width / 2.0f)) * m_scale.z;

	meshMinZ = (m_parentNode->positionZ - (m_parentNode->width / 2.0f)) * m_scale.x;
	meshMaxZ =( m_parentNode->positionZ + (m_parentNode->width / 2.0f)) * m_scale.z;


	if ((positionX <meshMinX) || (positionX > meshMaxX) || (positionZ < meshMinZ) || (positionZ > meshMaxZ))
	{
		return false;
	}

	FindNode(m_parentNode, positionX, positionZ, height);

	return true;
}


bool TerrainModel::GetNormalAtPosition(float positionX, float positionZ, Vector3& normal)
{
	float meshMinX, meshMaxX, meshMinZ, meshMaxZ;

	meshMinX = (m_parentNode->positionX - (m_parentNode->width / 2.0f)) * m_scale.x;
	meshMaxX = (m_parentNode->positionX + (m_parentNode->width / 2.0f)) * m_scale.z;

	meshMinZ = (m_parentNode->positionZ - (m_parentNode->width / 2.0f)) * m_scale.x;
	meshMaxZ = (m_parentNode->positionZ + (m_parentNode->width / 2.0f)) * m_scale.z;


	if ((positionX <meshMinX) || (positionX > meshMaxX) || (positionZ < meshMinZ) || (positionZ > meshMaxZ))
	{
		return false;
	}

	FindNodeNormal(m_parentNode, positionX, positionZ, normal);

	return true;
}

void TerrainModel::FindNode(NodeType* node, float x, float z, float& height)
{
	float xMin, xMax, zMin, zMax;
	int count, index;
	float vertex1[3], vertex2[3], vertex3[3];
	bool foundHeight;

	xMin = (node->positionX - (node->width / 2.0f)) * m_scale.x;
	xMax = (node->positionX + (node->width / 2.0f)) * m_scale.x;

	zMin = (node->positionZ - (node->width / 2.0f)) * m_scale.z;
	zMax = (node->positionZ + (node->width / 2.0f)) * m_scale.z;

	if ((x<xMin) || (x>xMax) || (z<zMin) || (z>zMax))
	{
		return;
	}

	count = 0;

	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i] != nullptr)
		{
			count++;
			FindNode(node->nodes[i], x, z, height);
		}
	}

	if (count > 0)
	{
		return;
	}

	for (int i = 0; i < node->triangleCount; i++)
	{
		index = i * 3;
		vertex1[0] = node->vertices[index].x;
		vertex1[1] = node->vertices[index].y;
		vertex1[2] = node->vertices[index].z;

		index++;
		vertex2[0] = node->vertices[index].x;
		vertex2[1] = node->vertices[index].y;
		vertex2[2] = node->vertices[index].z;

		index++;
		vertex3[0] = node->vertices[index].x;
		vertex3[1] = node->vertices[index].y;
		vertex3[2] = node->vertices[index].z;

		foundHeight = CheckHeightOfTriangle(x, z, height, vertex1, vertex2, vertex3);

		if (foundHeight)
		{
			return;
		}
	}
}

void TerrainModel::FindNodeNormal(NodeType* node, float x, float z, Vector3& normal)
{
	float xMin, xMax, zMin, zMax;
	int count, index;
	float vertex1[3], vertex2[3], vertex3[3];
	bool foundHeight;
	float height;

	xMin = (node->positionX - (node->width / 2.0f)) * m_scale.x;
	xMax = (node->positionX + (node->width / 2.0f)) * m_scale.x;

	zMin = (node->positionZ - (node->width / 2.0f)) * m_scale.z;
	zMax = (node->positionZ + (node->width / 2.0f)) * m_scale.z;

	if ((x<xMin) || (x>xMax) || (z<zMin) || (z>zMax))
	{
		return;
	}

	count = 0;

	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i] != nullptr)
		{
			count++;
			FindNodeNormal(node->nodes[i], x, z, normal);
		}
	}

	if (count > 0)
	{
		return;
	}

	for (int i = 0; i < node->triangleCount; i++)
	{
		index = i * 3;
		vertex1[0] = node->vertices[index].x;
		vertex1[1] = node->vertices[index].y;
		vertex1[2] = node->vertices[index].z;

		index++;
		vertex2[0] = node->vertices[index].x;
		vertex2[1] = node->vertices[index].y;
		vertex2[2] = node->vertices[index].z;

		index++;
		vertex3[0] = node->vertices[index].x;
		vertex3[1] = node->vertices[index].y;
		vertex3[2] = node->vertices[index].z;

		foundHeight = CheckHeightOfTriangle(x, z, height, vertex1, vertex2, vertex3);

		if (foundHeight)
		{
			normal.x = node->vertices[index].nx;
			normal.y = node->vertices[index].ny;
			normal.z = node->vertices[index].nz;
			return;
		}
	}
}

bool TerrainModel::CheckHeightOfTriangle(float x, float z, float& height, float v0[3], float v1[3], float v2[3])
{
	Vector3 startVector, directionVector, edge1, edge2, normal;
	Vector3 Q, e1, e2, e3, edgeNormal, temp;
	float magnitude, D, denominator, numerator, t, determinant;

	startVector.x = x;
	startVector.y = 0.0f;
	startVector.z = z;

	directionVector.x = 0.0f;
	directionVector.y = -1.0f;
	directionVector.z = 0.0f;

	edge1.x = v1[0] - v0[0];
	edge1.y = v1[1] - v0[1];
	edge1.z = v1[2] - v0[2];

	edge2.x = v2[0] - v0[0];
	edge2.y = v2[1] - v0[1];
	edge2.z = v2[2] - v0[2];

	normal.x = (edge1.y * edge2.z) - (edge1.z * edge2.y);
	normal.y = (edge1.z * edge2.x) - (edge1.x * edge2.z);
	normal.z = (edge1.x * edge2.y) - (edge1.y * edge2.x);

	normal.normalise();

	D = ((-normal.x * v0[0]) + (-normal.y * v0[1]) + (-normal.z * v0[2]));

	denominator = ((normal.x * directionVector.x) + (normal.y * directionVector.y) + (normal.z * directionVector.z));

	if (fabs(denominator) < 0.0001f)
	{
		return false;
	}

	numerator = -1.0f * (((normal.x * startVector.x) + (normal.y * startVector.y) + (normal.z * startVector.z)) + D);

	t = numerator / denominator;

	Q.x = startVector.x + (directionVector.x * t);
	Q.y = startVector.y + (directionVector.y * t);
	Q.z = startVector.z + (directionVector.z * t);

	e1.x = v1[0] - v0[0];
	e1.y = v1[1] - v0[1];
	e1.z = v1[2] - v0[2];

	e2.x = v2[0] - v1[0];
	e2.y = v2[1] - v1[1];
	e2.z = v2[2] - v1[2];

	e3.x = v0[0] - v2[0];
	e3.y = v0[1] - v2[1];
	e3.z = v0[2] - v2[2];

	edgeNormal.x = (e1.y * normal.z) - (e1.z * normal.y);
	edgeNormal.y = (e1.z * normal.x) - (e1.x * normal.z);
	edgeNormal.z = (e1.x * normal.y) - (e1.y * normal.x);

	temp.x = Q.x - v0[0];
	temp.y = Q.y - v0[1];
	temp.z = Q.z - v0[2];

	determinant = ((edgeNormal.x * temp.x) + (edgeNormal.y * temp.y) + (edgeNormal.z * temp.z));

	if (determinant > 0.001f)
	{
		return false;
	}

	edgeNormal.x = (e2.y * normal.z) - (e2.z * normal.y);
	edgeNormal.y = (e2.z * normal.x) - (e2.x * normal.z);
	edgeNormal.z = (e2.x * normal.y) - (e2.y * normal.x);

	temp.x = Q.x - v1[0];
	temp.y = Q.y - v1[1];
	temp.z = Q.z - v1[2];

	determinant = ((edgeNormal.x * temp.x) + (edgeNormal.y * temp.y) + (edgeNormal.z * temp.z));

	if (determinant > 0.001f)
	{
		return false;
	}

	edgeNormal.x = (e3.y * normal.z) - (e3.z * normal.y);
	edgeNormal.y = (e3.z * normal.x) - (e3.x * normal.z);
	edgeNormal.z = (e3.x * normal.y) - (e3.y * normal.x);

	temp.x = Q.x - v2[0];
	temp.y = Q.y - v2[1];
	temp.z = Q.z - v2[2];

	determinant = ((edgeNormal.x * temp.x) + (edgeNormal.y * temp.y) + (edgeNormal.z * temp.z));

	if (determinant > 0.001f)
	{
		return false;
	}

	height = Q.y + m_translate.y;

	return true;
}

void TerrainModel::CalculateNormals()
{
	int index1, index2, index3, index, count;
	Vector3 vertex1, vertex2, vertex3, vector1, vector2, sum, length;
	Vector3* normals;

	//Create a temporary array to hold normal vectors
	normals = new Vector3[(m_terrainHeight - 1) * (m_terrainWidth - 1)];

	for (int j = 0; j < (m_terrainHeight - 1); j++)
	{
		for (int i = 0; i < (m_terrainWidth - 1); i++)
		{
			index1 = ((j + 1) * m_terrainWidth) + i;
			index2 = ((j + 1) * m_terrainWidth) + (i + 1);
			index3 = (j * m_terrainWidth) + i;

			vertex1.x = m_heightMap[index1].x;
			vertex1.y = m_heightMap[index1].y;
			vertex1.z = m_heightMap[index1].z;

			vertex2.x = m_heightMap[index2].x;
			vertex2.y = m_heightMap[index2].y;
			vertex2.z = m_heightMap[index2].z;

			vertex3.x = m_heightMap[index3].x;
			vertex3.y = m_heightMap[index3].y;
			vertex3.z = m_heightMap[index3].z;

			vector1 = vertex1 - vertex3;
			vector2 = vertex3 - vertex2;

			index = (j * (m_terrainWidth - 1)) + i;

			normals[index] = vector1.cross(vector2);

			normals[index].normalise();
		}
	}

	for (int j = 0; j < m_terrainHeight; j++)
	{
		for (int i = 0; i < m_terrainWidth; i++)
		{
			sum.x = 0.0f;
			sum.y = 0.0f;
			sum.z = 0.0f;

			//bottom left
			if (((i - 1) >= 0) && ((j - 1) >= 0))
			{
				index = ((j - 1) * (m_terrainWidth - 1)) + (i - 1);
				sum += normals[index];
			}

			//bottom right
			if ((i < (m_terrainWidth - 1)) && ((j - 1) >= 0))
			{
				index = ((j - 1) * (m_terrainWidth - 1)) + i;
				sum += normals[index];
			}

			//upper left
			if (((i - 1) >= 0) && (j < (m_terrainHeight - 1)))
			{
				index = (j*(m_terrainWidth - 1)) + i;
				sum += normals[index];
			}

			//upper right
			if ((i < (m_terrainWidth - 1)) && (j < (m_terrainHeight - 1)))
			{
				index = (j * (m_terrainWidth - 1)) + i;
				sum += normals[index];
			}

			index = (j*m_terrainWidth) + i;
			sum.normalise();
			m_heightMap[index].nx = sum.x;
			m_heightMap[index].ny = sum.y;
			m_heightMap[index].nz = sum.z;
		}
	}

	delete[] normals;
}